/*
**
**  WAY (Where Are You ?)
**
**  parse.y		parser module.
**
**  Author: Roberto Bagnara (bagnara@vxcern.cern.ch),
**          CERN  DD Division OC Group RA Section.
**
**  All rights reserved to:
**
**          CERN
**          European Laboratory for Particle Physics
**          CH 1211 Geneva 23
**          SWITZERLAND
**
**  Permission is granted to any individual or institution
**  to use, copy, or redistribute this software so long as
**  it  is not  sold for  profit, provided  this copyright
**  notice is retained.
**
**  Modification History:
**
**  Version  Date    Who              Comments
**
**  1.0.00   890815  Roberto Bagnara  First working version
**
**
**  Description:
**
**  this module contains name server parser module. It is intended to be
**  processed by BISON, the GNU replacement for YACC.
**
*/


%{

/*
**      Include files.
**      --------------
*/

#include "way.h"
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include "parse.h"

/*
**      Forward references.
**      -------------------
*/

static int yylex();
static void yyerror();

/*
**	Flag to handle context-dependency, info fields are parsed differently.
**	----------------------------------------------------------------------
*/

static int parsing_info = 0;

/*
**      Variables to control parser input and actions.
**      ----------------------------------------------
*/

#ifdef __STDC__
static int  (*parser_input)(void);
static void (*parser_action)(NS_DECLARATION_PTR ns_decl);
#else
static int  (*parser_input)();
static void (*parser_action)();
#endif

/*
**      External functions.
**      -------------------
*/

#ifdef __STDC__
extern char *xmalloc(int size);
extern char *xrealloc(char *ptr, int size);
extern int keyword_type(const char *keyword);
#else
extern char *xmalloc();
extern char *xrealloc();
extern int keyword_type();
#endif

/*
**      External functions that don't like too much checking.
**      -----------------------------------------------------
*/

extern void fatal(/* function, format, arg1, arg2, ... */);
extern void error(/* function, format, arg1, arg2, ... */);
extern void xfree(/* any-pointer */);

%}


%start ns_file

/*
**	Semantic values.
**	----------------
*/

%union {
    char *           name_type;
    ADDRESS_PTR      address_type;
    SERVICE_LIST_PTR service_type;
    INFO_LIST_PTR    info_type;
}

%type <address_type> address
%type <address_type> address_field
%type <service_type> service
%type <service_type> service_field
%type <service_type> service_field_list
%type <info_type>    info
%type <info_type>    info_list
%type <name_type>    string

/*
**	Tokens provided by yylex().
**	---------------------------
*/

%token PROVIDES
%token <name_type> NAME
%token <name_type> FLAG
%token <name_type> AMBIGUOUS
%token <name_type> UNKNOWN
%token <name_type> NUMERIC
%token <name_type> ALPHA
%token <name_type> NUMBER


%%

/*
**	A "name server file" is either an empty file or a "name server file"
**	followed by a "name server declaration".
*/

ns_file : /* empty */
	| ns_file ns_decl
	;

/*
**	A "name server declaration" consists of an "address field" followed
**	by "::" followed by a "list of service fields".
*/

ns_decl : address_field PROVIDES service_field_list ';'
		{
		    NS_DECLARATION_PTR tmp;
		    register SERVICE_LIST_PTR sscan, stmp;
		    register INFO_LIST_PTR iscan, itmp;

		    /* Build the final structure. */

		    tmp = (NS_DECLARATION_PTR) xmalloc(sizeof(NS_DECLARATION));
		    tmp->address = $1;
		    tmp->service_list = $3;

		    /* Pass the whole structure to user function. */

		    (*parser_action)(tmp);

		    /* Release memory space: this looks quite complicated
		       and I wont even try to clarify it with comments.
		       Use of alloca() would simplify things a lot, but I
		       don't know how it works on all the funny machines
		       around here (PC, Mac, m68k systems for which only
		       brain damaged C libraries are provided etc.)
		    */

		    xfree(tmp->address->name);
		    iscan = tmp->address->info;
		    while (iscan) {
			xfree(iscan->name);
			itmp = iscan;
			iscan = iscan->next;
			xfree(itmp);
		    }
		    xfree(tmp->address);
		    sscan = tmp->service_list;
		    while (sscan) {
			xfree(sscan->name);
			iscan = sscan->info;
			while (iscan) {
			    xfree(iscan->name);
			    itmp = iscan;
			    iscan = iscan->next;
			    xfree(itmp);
			}
			stmp = sscan;
			sscan = sscan->next;
			xfree(stmp);
		    }
		    xfree(tmp);
		}
	;

/*
**	An "address field" is either an "address" or an "address" provided
**	with an "additional information list".
*/

address_field : address
		{
		    $1->info = NULL_INFO_LIST_PTR;
		    $$ = $1;
		}
	      | address '(' { parsing_info = 1; } info_list ')'
		{
		    parsing_info = 0;
		    $1->info = $4;
		    $$ = $1;
		}
	      ;

/*
**	A "service field list" is a comma-separated list of "service fields".
*/

service_field_list : service_field
		{
		    $$ = $1;
		}
		   | service_field_list ',' service_field
		{
		    $3->next = $1;
		    $$ = $3;
		}
		   ;

/*
**	A "service field" is either a "service" or a "service" provided
**	with an "additional information list".
*/

service_field : service
		{
		    $1->info = NULL_INFO_LIST_PTR;
		    $$ = $1;
		}
	      | service '(' { parsing_info = 1; } info_list ')'
		{
		    parsing_info = 0;
		    $1->info = $4;
		    $$ = $1;
		}
	      ;

/*
**	An "additional informations list" is a comma-separated list
**	additional informations.
*/

info_list : info
		{
		    $$ = $1;
		}
	      | info_list ',' info
		{
		    $3->next = $1;
		    $$ = $3;
		}
	      ;

address : NAME
		{
		    ADDRESS_PTR tmp;

		    tmp = (ADDRESS_PTR) xmalloc(sizeof(ADDRESS_ITEM));
		    tmp->name = $1;
		    $$ = tmp;
		}
	;

service : NAME
		{
		    SERVICE_LIST_PTR tmp;
	
		    tmp = (SERVICE_LIST_PTR) xmalloc(sizeof(SERVICE_LIST_ITEM));
		    tmp->name = $1;
		    tmp->next = NULL_SERVICE_LIST_PTR;
		    $$ = tmp;
		}
	;

info	: FLAG
		{
		    INFO_LIST_PTR tmp;

		    tmp = (INFO_LIST_PTR) xmalloc(sizeof(INFO_LIST_ITEM));
		    tmp->name = $1;
		    tmp->next = NULL_INFO_LIST_PTR;
		    $$ = tmp;
		}
	| NUMERIC '=' NUMBER
		{
		    INFO_LIST_PTR tmp;

		    tmp = (INFO_LIST_PTR) xmalloc(sizeof(INFO_LIST_ITEM));
		    tmp->name = $1;
		    tmp->value = $3;
		    tmp->next = NULL_INFO_LIST_PTR;
		    $$ = tmp;
		}
	| ALPHA '=' string
		{
		    INFO_LIST_PTR tmp;

		    tmp = (INFO_LIST_PTR) xmalloc(sizeof(INFO_LIST_ITEM));
		    tmp->name = $1;
		    tmp->value = $3;
		    tmp->next = NULL_INFO_LIST_PTR;
		    $$ = tmp;
		}
	;

string	: NAME
	| FLAG
	| AMBIGUOUS
	| UNKNOWN
	| NUMERIC
	| ALPHA
	| NUMBER
	;

%%


/*
**	Macros and variables for yylex lookahead mechanism.
**	---------------------------------------------------
*/

#define GETC      (cbufp == cbuf ? (*parser_input)() : *--cbufp)
#define UNGETC(c) (*cbufp++ = (c))

static int *cbufp;
static int cbuf[16];

/*
**	Just to provide a slightly better diagnostics.
**	----------------------------------------------
*/

static int line_number;


/*
**	The parser initialization function.
**	-----------------------------------
*/

#ifdef __STDC__
void init_parser(int (*input)(void), void (*action)(NS_DECLARATION_PTR decl))
#else
void init_parser(input, action)
int (*input)();
void (*action)();
#endif
{
    /* Initialize parser input and action function variables. */

    parser_input = input;
    parser_action = action;

    /* Initialize the line number counter. */

    line_number = 1;

    /* Initialize the yylex lookahead variables. */

    cbufp = cbuf;
}

/*
**	The parser error reporting function.
**	------------------------------------
*/

static void yyerror(string)
char *string;
{
    fatal("yyparse", "%s in line %d", string, line_number);
}


/*
**	The parser lexical analyzer.
**	----------------------------
*/

static char *name_terminators = "(),; \t\f\r\b\n/-=";

#define IS_TERMINATOR(c) (!(strchr(name_terminators, c) == (char *) 0))

static int yylex()
{
    int name_buf_len;
    char *name_buf;
    register int c, c1;
    register int i;
    int inside_comment;
    int is_quoted      = 0;
    int is_a_number    = 0;
    int is_hexadecimal = 0;

    c = GETC;
    while (1) {
	switch (c) {
	    case ' ' :			/* Skip all the white stuff. */
	    case '\t':
	    case '\f':
	    case '\r':
	    case '\b':
		c = GETC;
		break;

	    case '\n':			/* Increment the line number. */
		line_number++;
		c = GETC;
		break;

	    case '"':
		is_quoted = 1;
					/* INTENTIONALLY FALL THROUGH */
	    case '\\':
		c = GETC;
		goto get_name;
		break;

	    case '/':			/* Maybe a start of comment. */
		c1 = GETC;
		if (c1 != '*') {
		    UNGETC(c1);
		    goto get_name;	/* Not a comment, get name. */
		}

		c = GETC;
		inside_comment = 1;	/* We are inside a comment now. */
		while (inside_comment) {/* Skip to its end. */
		    if (c == '*') {
			while (c == '*')
			    c = GETC;

			if (c == '/') {	/* We are no longer inside it. */
			    inside_comment = 0;
			    c = GETC;
			}
		    } else if (c == '\n') {
			line_number++;	/* Continue taking care of line no. */
			c = GETC;
		    } else if (c == EOF) {
			error("yylex",
				"unexpected EOF inside comment at line %d",
				line_number);
			break;
		    } else
			c = GETC;	/* Continue skipping. */
		}
		break;

	    case '-':			/* Maybe the PROVIDES token. */
		c1 = GETC;
		if (c1 != '>') {
		    UNGETC(c1);
		    goto get_name;	/* No, it isn't, get name. */
		}

		return PROVIDES;	/* Yes, it is. */

	    case '(':			/* These are token themselves. */
	    case ')':
	    case ',':
	    case ';':
		return c;

	    case '=':			/* If we are parsing info */
		if (parsing_info)	/*  this is a token.      */
		    return c;
		else
		    goto get_name;
		break;

	    case EOF:			/* End of file. */
		return 0;

get_name:
	    default:

		/* Set the initial buffer length to 16. */

		name_buf_len = 16;
		name_buf = xmalloc(name_buf_len + 1);

		/* If we are parsing info we must check if we get a number. */

		if (parsing_info) {
		    is_a_number = 1;
		    is_hexadecimal = 0;
		}

		i = 0;
carry_on:
		do {

		    /* If buffer is full, make it bigger. */

		    if (i == name_buf_len) {
			name_buf_len *= 2;
			name_buf = xrealloc(name_buf, name_buf_len + 1);
		    }

		    /* Check if it is good for a number. */

		    if (parsing_info && is_a_number && !isdigit(c))
			switch (i) {
			    case 0:
				if (c != '-')
				    is_a_number = 0;
				break;
			    case 1:
				if ((c == 'x' || c == 'X') &&
							    name_buf[0] == '0')
				    is_hexadecimal = 1;
				else
				    is_a_number = 0;
				break;
			    default:
				if (is_hexadecimal &&
					strchr("abcdefABCDEF", c) == (char *) 0)
				    is_a_number = 0;
			}    /* switch */

		    /* Add current character to buffer. */

		    name_buf[i++] = c;

		    /* Get next character, if it is a back-slash
		       or we are inside a quoted string the normal
		       termination conditions don't apply.
		    */

		    c = GETC;
		    if (c == '\\') {
			c = GETC;
			continue;
		    }
		    if (is_quoted)
			if (c == EOF)
			    error("yylex",
			       "unexpected EOF inside quoted string at line %d",
				line_number);
			else if (c != '"')
			    continue;
			else
			    break;
		} while (c != EOF && !IS_TERMINATOR(c));

		/* Make sure it was a real terminator, that is one of
		   normal terminators or a comment delimiter or a
		   PROVIDES token or, if we are parsing the additiona
		   informations, an equal sign.  If not a real terminator
		   then carry on reading the name.
		*/

		if (c == '/' || c == '-') {
		    c1 = GETC;
		    UNGETC(c1);
		    if (c == '/' && c1 != '*' || c == '-' && c1 != '>')
			goto carry_on;
		}
		else if (c == '=' && !parsing_info)
		    goto carry_on;

		/* If it is not a quoted string put back the terminator. */

		if (!is_quoted)
		    UNGETC(c);

		/* Null-terminate the name buffer and return it
		   trying to guess the appropriate token.
		*/

		name_buf[i] = '\0';
		yylval.name_type = name_buf;

		if (parsing_info) {
		    if (is_a_number)
			return NUMBER;
		    if (!is_quoted)
			return keyword_type(name_buf);
		}
		return NAME;
	    }	/* switch */
    }		/* while */
}
